!*  <<< SUBINTERP >>> -------------------------------------------------
!======================================================================
!+

!     *****   2D IMAGE INTERPOLATION ROUTINES       *****

!     This file contains a set of functions to be used for
!     interpolating between image pixels, to find the value
!     of the image at fractional pixel locations.
!
!       FUNCTIONS:   BINSET     SET UP THE INTERPOLATION PROTOCOL
!                               FOR THE OTHER FUNCTIONS
!                    SNCBIN     SETS UP 2 1D SINC ARRAYS FOR REPEATED
!                               ROW AND COLUMN INTERPOLATION
!                    BINCOL     JUST INTERPOLATE ALONG IMAGE COLUMNS
!                    BINROW     JUST INTERPOLATE ALONG ROWS
!                    BIN        USE 2D BILINEAR INTERPOLATION
!                    XBIN       USE HIGH ACCURACY 2D SINC INTERPOLATION
!                    OUTBIN     USE LOWER (BUT FASTER) SINC FUNCTION
!                    AVRBIN     USE GAUSSIAN SMOOTHING
!  
!       AUTHOR: TOD R. LAUER   5/5/83   LICK OBSERVATORY
!               based on routines developed by Steve Kent at MIT
!       MODIFIED BY MARK BEDUZ             DAO
!       MODIFIED BY MASARU HAMABE          DAO/KISO
!       MODIFIED BY MASARU HAMABE          KISO   1988-06-28
!----------------------------------------------------------------------
!
!======================================================================

!*BINSET   -- set up a link for interpolation of data
!+
FUNCTION BINSET(gdata,JEXT,IEXT,DDLINE)
  parameter   (N2=17, N=N2/2, BETA=.30)
  parameter   (NO2=9, NO=NO2/2, BETA1=.60)
  real(4)   :: AR(N2), AC(N2), gdata(IEXT,JEXT)
  integer   :: IWT(-4:4,-4:4,4)
!
! BINSET:
!       This routine sets up a link for interpolation of data
!
! Arguments:
!         INPUT/
!           gdata  : THE IMAGE PIXEL ARRAY
!           JEXT   : THE NUMBER OF ROWS
!           IEXT   : THE NUMBER OF COLUMNS
!               THESE parameterS DEFINE THE 2D AND 1D SINC ARRAYS
!----------------------------------------------------------------------

  BINSET = FLOAT(N2)
  DLINE  = DDLINE
!     FOLLOWING STATEMENTS WERE ADDED BY M.HAMABE
!     CREATE GAUSSIAN WEIGHT PATTERNS
!       SG      : SIGMA IN PIXEL UNIT
!       IWT     : WEIGHT PATTERN
  do K=1,4
     SG=0.5*FLOAT(K)
     HH=1.0/(2.0*SG**2)
     do JJ=-4,4
        do II=-4,4
           RR=II*II+JJ*JJ
           IWT(II,JJ,K)=EXP(-HH*RR)*5.0+0.5
        end do
     end do
  end do
  return

!======================================================================

!*SNCBIN   -- set up 2 1D sinc arrays for later use by BINCOL & BINROW
!+
  ENTRY SNCBIN(DX,DY)

! SNCBIN:
!     This entry point sets up 2 1d sinc arrays for later use by bincol
!     and binrow
!
! Arguments:
!         INPUT/
!           DX     : FRACTIONAL SHIFT IN COLUMNS
!           DY     : FRACTIONAL SHIFT IN ROWS
!----------------------------------------------------------------------

  AR(1: N2) = 1.0
  AC(1: N2) = 1.0
  IX = NINT(DX)
  IY = NINT(DY)
  DX = DX-IX
  DY = DY-IY
  if ( DX>0.5 ) DX = DX-1.0
  if ( DY>0.5 ) DY = DY-1.0
  SINDX = -1**N
  SINDY = -1**N
  if ( DX<0.0 ) SINDX = -SINDX
  if ( DY<0.0 ) SINDY = -SINDY
  DXN = -N-DX
  DYN = -N-DY
!     GENERATE THE SINC ARRAYS WITH EXPONENTIAL TAPERING
  do I=1, N2
     AX = I+DXN
     AY = I+DYN
     if ( AX==0.0 ) then
        PX = 1.0
     else
        PX = SINDX/AX*EXP(-(AX*BETA)**2)
     end if
     if ( AY==0.0 ) then
        PY = 1.0
     else
        PY = SINDY/AY*EXP(-(AY*BETA)**2)
     end if
     AR(I) = AR(I)*PY
     AC(I) = AC(I)*PX
     SINDX = -SINDX
     SINDY = -SINDY
  end do
!     NORMALIZE THE ARRAYS TO UNITY.
  SUMR = 0.0
  SUMC = 0.0
  do I=1, N2
     SUMR = SUMR+AR(I)
     SUMC = SUMC+AC(I)
  end do
  DIV = FLOAT(N2)
  SUMR = SUMR/DIV
  SUMC = SUMC/DIV
  do I=1,N2
     AR(I) = AR(I)/SUMR
     AC(I) = AC(I)/SUMC
  end do
  
  return
!
!======================================================================

!*BINCOL   -- get image value at new column location
!+
  ENTRY BINCOL(X,Y,gdata,JEXT,IEXT)
!
! BINCOL:
!     Get image value at new column location
!     This entry point assumes an already calculated sinc array.   It
!     is used for repeated interpolation of the same fractional amount.
!
! Arguments:
!         INPUT/
!           X      : COLUMN LOCATION OF NEW PIXEL
!           Y      : ROW LOCATION OF NEW PIXEL
!
!----------------------------------------------------------------------
  BINCOL = 0.0
  IX = NINT(X)
  IY = NINT(Y)
  if ( IY<1   ) IY = 1
  if ( IY>JEXT) IY = JEXT
!!!!  if ( DX>0.5 ) then    DX is not defined
!!!!     IX = IX+1
!!!!  end if
  IXN = IX-N-1
!     CONVOLVE THE IMAGE ROW WITH THE 1D SINC ARRAY, AND SUM TO GET
!     THE NEW PIXEL VALUE.
!     IF AT THE IMAGE EDGE, JUST REPEAT THE EDGE PIXEL.
  do I=1, N2
     NX = MIN(IEXT,MAX(1,IXN+I))
     BINCOL = BINCOL+AC(I)*gdata(NX,IY)
  end do
  BINCOL = BINCOL/DIV
                                                                        
  return
!
!======================================================================
!*BINROW   -- get image value at new row location
!+
  ENTRY BINROW(X,Y,gdata,JEXT,IEXT)
!
! BINROW:
!     Get image value at new row location
!     This entry point assumes an already calculated sinc array. It
!     is used for repeated interpolation of the same fractional amount.
!
! Arguments:
!         INPUT/
!           X      : COLUMN LOCATION OF NEW PIXEL
!           Y      : ROW LOCATION OF NEW PIXEL
!
!----------------------------------------------------------------------
  BINROW = 0.0
  IX = NINT(X)
  IY = NINT(Y)
  if ( IX<1    ) IX = 1
  if ( IX>IEXT ) IX = IEXT
!!!  if ( DY>0.5  ) then   DY is not defined
!!!     IY = IY+1
!!!  end if
  IYN = IY-N-1
!     CONVOLVE THE IMAGE ROW WITH THE 1D SINC ARRAY, AND SUM TO GET
!     THE NEW PIXEL VALUE.
!     IF AT THE IMAGE EDGE, JUST REPEAT THE EDGE PIXEL.
  do I=1,N2
     NY = MIN(JEXT,MAX(1,IYN+I))
     BINROW = BINROW+AR(I)*gdata(IX,NY)
  end do

  BINROW = BINROW/DIV
  return

!======================================================================

!*BIN      -- simple bilinear interpolation in an image arry
!+
  ENTRY BIN(X,Y,gdata,JEXT,IEXT)
!
! BIN:
!     Compute interpolated value in image array at point (x,y).
!     This entry point uses a simple bilinear interpolation scheme,
!     Linear interpolaton between the 4 pixels surrounding the x, y
!     position.
!
!--
!     MODIFIED TO return A VALUE OF -100000 INSTEAD OF 0 WHEN OUTSIDE
!     THE LEGAL RANGE OF POINTS.                              86/01/16
!----------------------------------------------------------------------
  BIN = DLINE
  IX = X
  IY = Y
  if ( IX<1 .or. IX+1>IEXT .or. IY<1 .or. IY+1>JEXT ) return
  if ( gdata(IX  ,IY)<=DLINE .or. gdata(IX  ,IY+1)<=DLINE .or. &
       & gdata(IX+1,IY)<=DLINE .or. gdata(IX+1,IY+1)<=DLINE ) return
  BIN =    gdata(IX  ,IY  )*(IY+1.-Y)*(IX+1.-X) &
       & + gdata(IX  ,IY+1)*(Y   -IY)*(IX+1.-X) &
       & + gdata(IX+1,IY+1)*(Y   -IY)*(X   -IX) &
       & + gdata(IX+1,IY  )*(IY+1.-Y)*(X   -IX)
  
  return

!======================================================================

!*XBIN     -- high accuracy 2d interpolation routine
!+
  ENTRY XBIN(X,Y,gdata,JEXT,IEXT)

! XBIN:
!     High accuracy 2d interpolation routine.
!     This entry point does not check for zones or negative pixels.
!     use sin(x)/x*exp(-x**2) form.
!
! Arguments:
!         INPUT/
!           X      : COLUMN LOCATION OF THE NEW PIXEL
!           Y      : ROW LOCATION
!
!--
!     MODIFIED TO CHECK THAT THE POINTS ARE NOT OUTSIDE THE IMAGE,
!     IN WHICH CASE A VALUE OF -100000 IS returnED.          86/01/16
!----------------------------------------------------------------------
!     print *, 'start XBIN'
!     print *, 'x,y,jext,iext,dline,n', x,y,jext,iext,dline,n
  XBIN = DLINE
  IX = int(X)
  IY = int(Y)
  if ( IX<1 .or. IX>IEXT .or. IY<1 .or. &
       &  IY>JEXT .or. gdata(IX,IY)<=DLINE ) return
!     print *, 'check0 XBIN'
  xdx = X-IX
  xdy = Y-IY
!     print *, 'check1 XBIN'
  if ( xdx>0.5 ) then
     xdx = xdx-1.0
     IX = IX+1
  end if
  if ( xdy>0.5 ) then
     xdy = xdy-1.0
     IY = IY+1
  end if
  SINDX = -1**N
  SINDY = -1**N
  if ( xdx<0.0 ) SINDX=-SINDX
  if ( xdy<0.0 ) SINDY=-SINDY
  DXN = -1-N-xdx
  DYN = -1-N-xdy
!     GENERATE A 2D SINC ARRAY WITH EXPONENTIAL TAPERING.  WE do NOT
!     NEED TO CALCULATE ANY SINE FUNCTIONS, SINCE WE LATER NORMALIZE
!     THE ARRAY ANYWAY.
!     print *, 'check2 XBIN'
  do I=1, N2
     AX = I+DXN
     AY = I+DYN
     if ( AX==0.0 ) then
        PX = 1.0
     else if ( xdx==0.0 ) then
        PX = 0.0
     else
        PX = SINDX/AX*EXP(-(AX*BETA)**2)
     end if
     if ( AY==0.0 ) then
        PY = 1.0
     else if ( xdy==0.0 ) then
        PY = 0.0
     else
        PY = SINDY/AY*EXP(-(AY*BETA)**2)
     end if
     AC(I) = PX
     AR(I) = PY
     SINDX = -SINDX
     SINDY = -SINDY
  end do
!     print *, 'check3 XBIN'
!     NORMALIZE THE ARRAY TO UNITY.
  SUMX = 0.0
  SUMY = 0.0
  do I=1,N2
     SUMX = SUMX+AC(I)
     SUMY = SUMY+AR(I)
  end do
!
!     CONVOLVE THE IMAGE ARRAY WITH THE SINC ARRAY AND SUM
!     TO GENERATE THE NEW PIXEL VALUE AT X AND Y.
!
!     print *, 'check4 XBIN'
  XBIN = 0.0
  IXN = IX-N-1
  IYN = IY-N-1
  do I=1,N2
     SUM = 0.0
     do J=1,N2
        NX = MIN(IEXT,MAX(1,IXN+J))
        NY = MIN(JEXT,MAX(1,IYN+I))
        SUM = SUM+AC(J)*gdata(NX,NY)
     end do
     XBIN = XBIN+AR(I)*SUM
  end do
!     print *, 'check5 XBIN'

  XBIN = XBIN/(SUMX*SUMY)
!     print *, 'end   XBIN'
  return

!======================================================================

!*OUTBIN   -- lower accuracy (but quicker) sinc interpolation routine
!+
  ENTRY OUTBIN(X,Y,gdata,JEXT,IEXT)
!
! OUTBIN:
!     Lower accuracy (but quicker) sinc interpolation routine.
!     This entry point does not check for zones or negative pixels.
!     Use sin(x)/x*exp(-x**2) form.
!
! Arguments:
!         INPUT/
!           X      : COLUMN LOCATION OF THE NEW PIXEL
!           Y      : ROW LOCATION
!
!-
!     MODIFIED TO CHECK THAT THE POINTS ARE NOT OUTSIDE THE IMAGE,
!     IN WHICH CASE A VALUE OF -100000 IS returnED.            86/01/16
!----------------------------------------------------------------------
  OUTBIN = DLINE
  IX = X
  IY = Y
  if ( IX<1 .or. IX>IEXT .or. IY<1 .or. &
       &  IY>JEXT .or. gdata(IX,IY)<=DLINE ) return
  odx = X-IX
  ody = Y-IY
  if ( odx>0.5 ) then
     odx = odx-1.0
     IX = IX+1
  end if
  if ( ody>0.5 ) then
     ody = ody-1.0
     IY = IY+1
  end if
  SINDX = -1**NO
  SINDY = -1**NO
  if ( odx<0.0 ) SINDX=-SINDX
  if ( ody<0.0 ) SINDY=-SINDY
  DXN = -1-NO-odx
  DYN = -1-NO-ody

!     GENERATE A 2D SINC ARRAY WITH EXPONENTIAL TAPERING.  WE do NOT
!     NEED TO CALCULATE ANY SINE FUNCTIONS, SINCE WE LATER NORMALIZE
!     THE ARRAY ANYWAY.

  do I=1, NO2
     AX = I+DXN
     AY = I+DYN
     if ( AX==0.0 ) then
        PX = 1.0
     else if ( odx==0.0 ) then
        PX = 0.0
     else
        PX = SINDX/AX*EXP(-(AX*BETA1)**2)
     end if
     if ( AY==0.0 ) then
        PY = 1.0
     else if ( ody==0.0 ) then
        PY = 0.0
     else
        PY = SINDY/AY*EXP(-(AY*BETA1)**2)
     end if
     AC(I) = PX
     AR(I) = PY
     SINDX = -SINDX
     SINDY = -SINDY
  end do

!     NORMALIZE THE ARRAY TO UNITY
  SUMX = 0.0
  SUMY = 0.0
  do I=1,NO2
     SUMX = SUMX+AC(I)
     SUMY = SUMY+AR(I)
  end do

!     CONVOLVE THE IMAGE ARRAY WITH THE SINC ARRAY AND SUM TO GENERATE
!     THE NEW PIXEL VALUE AT X AND Y.
  OUTBIN = 0.0
  IXN = IX-NO-1
  IYN = IY-NO-1
  do I=1,NO2
     SUM = 0.0
     do J=1,NO2
        NX = MIN(IEXT,MAX(1,IXN+J))
        NY = MIN(JEXT,MAX(1,IYN+I))
        SUM = SUM+AC(J)*gdata(NX,NY)
     end do
     OUTBIN = OUTBIN+AR(I)*SUM
  end do
!
  OUTBIN = OUTBIN/(SUMX*SUMY)
  return
!======================================================================

!*AVRBIN   -- compute averaged value at a point (x, y) in an image array.
!+
  ENTRY AVRBIN(X,Y,ISGM,gdata,JEXT,IEXT)
!
! AVRBIN:
!     Compute averaged value in image array at point (x,y).
!     This entry point uses an average of pixels around the point
!     (x,y). The average is calculated using the gaussian-weight
!     patterns with shigmas of 0.5, 1.0, 1.5, 2.0 (pixel).
!       sigma=isgm*0.5pixel

!     See also function BINSET
!
!--
!     AUTHOR: MASARU HAMABE   JAN 26, 1986   D.A.O.(KISO OBS.,T.A.O.)
!             BASED ON ROUTINE IN SPIRAL AT KISO OBS.
!----------------------------------------------------------------------
  AVRBIN = DLINE
  IX = X+0.5
  IY = Y+0.5
  if ( IX<1 .or. IX>IEXT .or. IY<1 .or. IY>JEXT ) return
  
  if ( ISGM<1 ) ISGM=1
  if ( ISGM>4 ) ISGM=4
  NP=0
  AVRBIN=0.0
  do J = -ISGM, ISGM
     JJ=IY+J
     if ( JJ>=1 .and. JJ<=JEXT ) then
        do I = -ISGM, ISGM
           II=IX+I
           if ( II>=1 .and. II<=IEXT .and. gdata(II,JJ)>DLINE ) then
              AVRBIN=AVRBIN+gdata(II,JJ)*FLOAT( IWT(I,J,ISGM) )
              NP=NP+IWT(I,J,ISGM)
           end if
        end do
     end if
  end do
  if ( NP>0 ) then
     AVRBIN = AVRBIN/FLOAT(NP)
  else
     AVRBIN = DLINE
  end if
  !
  return
END FUNCTION BINSET

!======================================================================

!*FMEDIAN  -- find the median in an array of numbers
!+ 
SUBROUTINE FMEDIAN(X,NPTS,FMEDIA)
  dimension  X(NPTS)
                                                                        
! FMEDIAN:
!     This routine finds the median in an array of numbers.
!     it uses a quicksort algorithm from the text :
!     'Algorithms + Data Structures = Programs' by Wirth (1976) P.84.
!     In the worst case it completely sorts the input array.
!     (time is proportional to n * log n.)
!     sorting is done on the indices of the array, so as not to
!     disturb its original ordering.

! Arguments:
!         INPUT/
!           X      : THE ARRAY OF POINTS TO BE EXAMINED
!           NPTS   : THE LENGTH OF THE ARRAY
!         OUTPUT/
!           FMEDIA : THE MEDIAN VALUE

!--
!     AUTHORS:MIKE FICH, BERKELEY, APRIL 1981.
!     B.F.
!---------------------------------------------------------------------
                          
  INTEGER :: KEY(1000)
!                KEY    : THE KEYS TO BE PARTIALLY SORTED
                          
!     INITIALIZE THE KEYS.
  do I=1, NPTS
     KEY(I) = I
  end do
  ILEFT = 1
  IRITE = NPTS
  MIDDLE = (NPTS+1)/2
                          
!     FIND THE MIDDLE ELEMENT, SORT EVERYTHING BY EXCHANGING EVERYTHING
!     BIGGER ON THE LEFT WITH EVERYTHING SMALLER ON THE RIGHT END OF
!     THE ARRAY.
                                                                        
  do while (ILEFT<IRITE)
     XMID = X(KEY(MIDDLE))
                                                                        
!    INDEX1 MOVES FROM LEFT TO RIGHT, INDEX2 MOVES FROM RIGHT TO LEFT
!    INDEX1 POINTS TO SOMETHING BIGGER THAN THE KEY BEING SORTED ON.
!    INDEX2 POINTS TO SOMETHING SMALLER.
                                                     
     INDEX1 = ILEFT
     INDEX2 = IRITE
     do while ( INDEX1<=INDEX2 )
        if ( X(KEY(INDEX1))<XMID ) then
           do while ( X(KEY(INDEX1))<XMID )
              INDEX1 = INDEX1+1
           end do
        end if
        if ( X(KEY(INDEX2))>XMID ) then
           do while ( X(KEY(INDEX2))>XMID )
              INDEX2 = INDEX2-1
           end do
        end if
        
!    IF INDEX1 IS STILL TO THE LEFT OF INDEX2 then EXCHANGE THE
!    KEYS FOUND OUT OF PLACE AND INCREMENT POINTERS.
                                                
        if ( INDEX1<=INDEX2 ) then
           KTEMP = KEY(INDEX1)
           KEY(INDEX1) = KEY(INDEX2)
           KEY(INDEX2) = KTEMP
           INDEX1 = INDEX1+1
           INDEX2 = INDEX2-1
        end if
        
!        IF NOT SORTED FOR INITIAL KEY GO BACK.
     end do
                                                
!        CHOOSE NEW LIMITS FOR SORTING REMAINING BLOCK.
     if ( INDEX2<MIDDLE ) ILEFT = INDEX1
     if ( MIDDLE<INDEX1 ) IRITE = INDEX2
  end do
  FMEDIA = X(KEY(MIDDLE))
  
  return

END SUBROUTINE FMEDIAN
